/* NOT FINISHED */

typedef struct {
	long stamp;
	long count;
	sfsistat status;
} MilterCacheEntry;

extern int MilterCacheOpen(const char *file);
extern int MilterCacheGet(MilterWork *data, const char *key, MilterCacheEntry *value);
extern int MilterCachePut(MilterWork *data, const char *key, MilterCacheEntry *value);
extern int MilterCacheGc(MilterWork *data, long frequency, long ttl);
extern int MilterCacheClose(void);

static volatile Cache milterCache;
static volatile long milterCacheGcCount;

static MilterCacheEntry milterCacheUndefinedEntry = { 0, 0, X_SMFIS_UNKNOWN };

/***********************************************************************
 *** Cache Support
 ***********************************************************************/

int
MilterCacheOpen(const char *type, const char *file)
{
	if (milterCache == NULL && (milterCache = CacheCreate(type, file)) == NULL) {
		fprintf(stderr, "failed to create cache\n");
		return 1;
	}
	
	return 0;
}

void
MilterCacheClose(void)
{
	milterCache->sync(milterCache);
	milterCache->destroy(milterCache);
}

int
MilterCacheGet(MilterWork * data, const char *name, MilterCacheEntry *entry)
{
	Data value;
	struct data key;

	*entry = milterCacheUndefinedEntry;	
	DataInitFromBytes(&key, name, strlen(name)+1);

	if (milterLogDetail & MILTER_LOG_CACHE)
		syslog(LOG_DEBUG, "cache get [%s]", name);

	if (pthread_mutex_lock(&milterMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex lock in MilterCacheGet() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	value = milterCache->get(milterCache, &key);

	if (pthread_mutex_unlock(&milterMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex unlock in MilterCacheGet() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	if (value == NULL)
		return -1;
	
	*entry = *(MilterCacheEntry *)(value->base(value));
	value->destroy(value);

	if (milterLogDetail & MILTER_LOG_CACHE)
		syslog(LOG_DEBUG, TAG_FORMAT "cacheGet(%lx, '%s', {st=%d, cn=%ld}) ", TAG_ARGS, (long) data, name, entry->status, entry->count);

	return 0;
}

int
MilterCachePut(MilterWork * data, const char *name, MilterCacheEntry *entry)
{
	int rc;
	struct data key, value;

	DataInitFromBytes(&key, name, strlen(name)+1);
	DataInitFromBytes(&value, (unsigned char *) entry, sizeof (*entry));

	if (milterLogDetail & MILTER_LOG_CACHE)
		syslog(LOG_DEBUG, "cache put [%s]", name);
	
	if (pthread_mutex_lock(&milterMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex lock in MilterCachePut() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);

	rc = milterCache->put(milterCache, &key, &value);
	
	if (pthread_mutex_unlock(&milterMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex unlock in MilterCachePut() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	return rc;
}

static int
MilterCacheExpireEntries(void *key, void *value, void *data)
{
	time_t *now = data;
	MilterCacheEntry *entry = (MilterCacheEntry *) ((Data) value)->base(value);
	
	if (*now < entry->stamp)
		return 1;

	if (milterLogDetail & MILTER_LOG_CACHE)
		syslog(LOG_DEBUG, "cache remove <%s>", ((Data) key)->base(key));

	return -1;	
}

int
MilterCacheGarbageCollect(MilterWork *data, long frequency, long ttl)
{
	time_t now = time(NULL) - ttl;
	
	if (pthread_mutex_lock(&milterMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex lock in MilterCacheGarbageCollect() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);		

	if (frequency <= ++milterCacheGcCount) {
		if (milterLogDetail & MILTER_LOG_CACHE)
			syslog(LOG_DEBUG, "%ld connections, garbage collecting cache", milterCacheGcCount);
			
		milterCache->walk(milterCache, MilterCacheExpireEntries, &now);
		milterCacheGcCount = 0;

		if (milterLogDetail & MILTER_LOG_CACHE)
			syslog(LOG_DEBUG, "syncing cache");
		milterCache->sync(milterCache);
	}

	if (pthread_mutex_unlock(&milterMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex unlock in MilterCacheGarbageCollect() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	return 0;
}

